#include "aes_core.h"
#include <assert.h>
#include <errno.h>

/* next ocb counter */
static inline block
ocb_double(block x, const aes_consts *c) {
  return xor(pslld(x,1),
	     and(c->ocb_mask,
		 psrad(pshufd(x,3,0,1,2), 31)));
}


/* increment a 128-bit counter */
 block
ctr_128_increment(block x, const aes_consts *c) {
   block high_side = and(pslldq(x,8), pshufd(x,0,0,1,0));
   return psubq(x, pcmpeqd(high_side, c->ctr_mask));
}

off_t
aes_encrypt(const byte      *input,
	    size_t           input_size,
	    byte            *output,
	    size_t           output_size,
	    const aes_key   *key,
	    const byte      *iv,
	    aes_mode         mode,
	    aes_padding      padding) {
  /* Sanity check mode. */
  switch(mode) {
  case AES_MODE_OFB:
  case AES_MODE_CTR_128:
  case AES_MODE_CTR_64:
    if (padding != AES_PADDING_NONE)
      return -EINVAL;
    else
      break;

  case AES_MODE_ECB:
  case AES_MODE_CBC:
  case AES_MODE_CFB:
    break;

  case AES_MODE_OCB:
    if (padding != AES_PADDING_NONE)
      return -EINVAL;
    else
      return -ENOSYS;

  default:
    return -EINVAL;
  };

  /* Sanity check padding and compute required output size. */
  off_t output_size_required;
  switch(padding) {
  case AES_CIPHERTEXT_STEALING:
    output_size_required = (off_t) input_size;
    if (input_size % block_size == 0)
      padding = AES_PADDING_NONE;
    else if (input_size < block_size)
      /* too short to steal ciphertext */
      return -EDOM;
    break;

  case AES_PADDING_NONE:
    if ((mode == AES_MODE_ECB || mode == AES_MODE_CBC)
	&& input_size % block_size != 0)
      return -EINVAL;
    output_size_required = (off_t) input_size;
    break;

  case AES_PADDING_NULL:
    if (input_size % block_size == 0) {
      padding = AES_PADDING_NONE;
      output_size_required = (off_t) input_size;
      break;
    }

  case AES_PADDING_DES:
  case AES_PADDING_PKCS5:
    output_size_required = (((off_t) input_size) + block_size)
      & ~(block_size - 1);
    break;

  default:
    return -EINVAL;
  }
  
  if (output_size_required < 0
      || output_size_required > output_size)
    return -E2BIG;

  /* fire it up */
  aes_use_locals();
  size_t nbytes = input_size;
  block chain = zero();

  switch(mode) {
  case AES_MODE_ECB: {
    /* ECB mode */
    for (; nbytes >= block_size;
	   input  += block_size,
	   output += block_size,
	   nbytes -= block_size) {
      storeu(output, chain = aes_encrypt_core(loadu(input), key));
    }

    switch(padding) {
    case AES_PADDING_NULL:
    case AES_PADDING_DES:
    case AES_PADDING_PKCS5:
      storeu(output, aes_encrypt_core(loadup(input, nbytes,
					     chain, padding),
				      key));
      break;

    case AES_CIPHERTEXT_STEALING: {
      block ipt2 = loadup(input, nbytes, chain, padding);
      storeup(output, nbytes, chain);
      storeu(output - block_size, aes_encrypt_core(ipt2, key));
      break;
    }

    case AES_PADDING_NONE:
      break;
    };

    goto done;
  };

  case AES_MODE_CTR_64: {
    /* CTR mode, 64+64-bit counter */
    chain = loadu(iv);
    for (; nbytes >= block_size;
	   input  += block_size,
	   output += block_size,
	   nbytes -= block_size) {
      storeu(output, xor(loadu(input),
			 aes_encrypt_core(chain, key)));
      chain = paddq(chain, _base->ctr_one);
    }

    goto finish_prf;
  }

  case AES_MODE_CTR_128: {
    /* CTR mode, 128-bit counter */
    chain = loadu(iv);

    for (; nbytes >= block_size;
	   input  += block_size,
	   output += block_size,
	   nbytes -= block_size) {
      storeu(output, xor(loadu(input),
			 aes_encrypt_core(chain, key)));
      chain = ctr_128_increment(chain, _base);
    }

    goto finish_prf;
  }

  case AES_MODE_CBC: {
    chain = loadu(iv);

    for (; nbytes >= block_size;
	   input  += block_size,
	   output += block_size,
	   nbytes -= block_size) {
      storeu(output, chain
	     = aes_encrypt_core(xor(chain, loadu(input)), key));
    }

    switch(padding) {
    case AES_PADDING_NULL:
    case AES_PADDING_DES:
    case AES_PADDING_PKCS5:
      storeu(output,
	     aes_encrypt_core(xor(chain,
				  loadup(input, nbytes,
					 chain, padding)),
			      key));
      break;

    case AES_CIPHERTEXT_STEALING: {
      block ipt2 = xor(chain,
		       loadup(input, nbytes,
			      chain, AES_PADDING_NULL));
      storeup(output, nbytes, chain);
      storeu(output - block_size, aes_encrypt_core(ipt2,key));
      break;
    }
      
    case AES_PADDING_NONE:
      break;
    };

    goto done;
  }

  case AES_MODE_OFB: {
    /* Output feedback mode */
    chain = loadu(iv);
    for (; nbytes >= block_size;
	   input  += block_size,
	   output += block_size,
	   nbytes -= block_size) {
      chain = aes_encrypt_core(chain, key);
      storeu(output, xor(loadu(input), chain));
    }

    goto finish_prf;
  }

  case AES_MODE_CFB: {
    /* Ciphertext feedback mode */
    chain = loadu(iv);
    for (; nbytes >= block_size;
	   input  += block_size,
	   output += block_size,
	   nbytes -= block_size) {
      storeu(output, chain =
	     xor(loadu(input), aes_encrypt_core(chain, key)));
    }

    goto finish_prf;
  }

  default:
    assert(0);
  };

  goto done;

 finish_prf:
  if (nbytes > 0) {
    storeup(output, nbytes,
	    xor(aes_encrypt_core(chain, key),
		loadup(input, nbytes,
		       chain, AES_PADDING_NULL)));
  }    

 done:
  aes_cleanup();
  return output_size_required;
}


off_t
aes_decrypt(const byte      *input,
	    size_t           input_size,
	    byte            *output,
	    size_t           output_size,
	    const aes_key   *key,
	    const byte      *iv,
	    aes_mode         mode,
	    aes_padding      padding) {
  byte *orig_output = output;

  /* Sanity check mode. */
  switch(mode) {
  case AES_MODE_CTR_128:
  case AES_MODE_CTR_64:
  case AES_MODE_OFB:
    return aes_encrypt(input, input_size,
		       output, output_size,
		       key, iv, mode, padding);

  case AES_MODE_ECB:
  case AES_MODE_CBC:
  case AES_MODE_CFB:
    break;

  case AES_MODE_OCB:
    if (padding != AES_PADDING_NONE)
      return -EINVAL;
    else
      return -ENOSYS;

  default:
    return -EINVAL;
  };

  /* compute require output size */
  off_t output_size_required;
  switch(padding) {
  case AES_CIPHERTEXT_STEALING:
    output_size_required = (off_t) input_size;
    if (input_size % block_size == 0)
      padding = AES_PADDING_NONE;
    else if (input_size < block_size)
      /* too short to steal ciphertext */
      return -EDOM;
    break;

  case AES_PADDING_NONE:
    if ((mode == AES_MODE_ECB || mode == AES_MODE_CBC)
	&& input_size % block_size != 0)
      return -EINVAL;
    output_size_required = (off_t) input_size;
    break;

  case AES_PADDING_PKCS5:
  case AES_PADDING_DES:
    if (input_size % block_size != 0)
      return -EINVAL;

    // hax
    output_size_required = input_size - 16;
    break;

  case AES_PADDING_NULL:
    if (input_size % block_size != 0)
      return -EINVAL;

    // hax
    output_size_required = input_size - 16;
    if (output_size_required == output_size)
      // not actually attainable
      return -E2BIG;
    break;

  default:
    return -EINVAL;
  }

  if (output_size_required < 0
      || output_size_required > output_size)
    return -E2BIG;

  /* fire it up */
  aes_use_locals();
  size_t nbytes = input_size;
  block chain = zero();

  switch(mode) {
  case AES_MODE_ECB: {
    /* ECB mode */
    /* note the >, for padding */
    for (; nbytes >  block_size;
	   input  += block_size,
	   output += block_size,
	   nbytes -= block_size) {
      storeu(output, chain
	     = aes_decrypt_core(loadu(input), key));
    }

    switch(padding) {
    case AES_PADDING_NULL:
    case AES_PADDING_DES:
    case AES_PADDING_PKCS5:
      assert(nbytes == block_size);
      chain = aes_decrypt_core(loadu(input), key);

      int npad = check_padding(chain, padding);
      assert(npad <= block_size);

      output_size_required += npad;
      if (npad < 0) {
	output_size_required = -EBADMSG;
      } else if (output_size_required > output_size) {
	output_size_required = -E2BIG;
      } else {
	storeup(output, npad, chain);
      }
      break;

    case AES_CIPHERTEXT_STEALING: {
      block ipt2 = loadup(input, nbytes, chain, padding);
      storeup(output, nbytes, chain);
      storeu(output - block_size, aes_decrypt_core(ipt2, key));
      break;
    }

    case AES_PADDING_NONE:
      storeu(output, aes_decrypt_core(loadu(input), key));
      break;
    };
    break;
  };

  case AES_MODE_CBC: {
    chain = loadu(iv);
    block last_pt = zero();

    for (; nbytes >  block_size;
	   input  += block_size,
	   output += block_size,
	   nbytes -= block_size) {
      block ct = loadu(input);
      last_pt = aes_decrypt_core(ct, key);
      storeu(output, xor(chain, last_pt));
      chain = ct;
    }

    switch(padding) {
    case AES_PADDING_NULL:
    case AES_PADDING_DES:
    case AES_PADDING_PKCS5:
      assert(nbytes == block_size);
      chain = xor(chain, aes_decrypt_core(loadu(input), key));

      int npad = check_padding(chain, padding);
      assert(npad <= block_size);

      output_size_required += npad;
      if (npad < 0) {
	output_size_required = -EBADMSG;
      } else if (output_size_required > output_size) {
	output_size_required = -E2BIG;
      } else {
	storeup(output, npad, chain);
      }
      break;

    case AES_CIPHERTEXT_STEALING: {
      block ct2 = loadup(input, nbytes, last_pt, padding);
      storeup(output, nbytes, xor(last_pt, ct2));
      storeu(output - block_size,
	     xor(loadu(output - block_size),
		 xor(last_pt,
		     aes_decrypt_core(ct2,key))));
      break;
    }
      
    case AES_PADDING_NONE:
      storeu(output, xor(chain, aes_decrypt_core(loadu(input), key)));
      break;
    };
    break;
  }

  case AES_MODE_CFB: {
    /* Ciphertext feedback mode */
    chain = loadu(iv);
    for (; nbytes >= block_size;
	   input  += block_size,
	   output += block_size,
	   nbytes -= block_size) {
      block ct = loadu(input);
      storeu(output, xor(ct, aes_encrypt_core(chain, key)));
      chain = ct;
    }

    if (nbytes > 0) {
      storeup(output, nbytes,
	      xor(aes_encrypt_core(chain, key),
		  loadup(input, nbytes,
			 chain, AES_PADDING_NULL)));
    }
    break;
  }

  default:
    break;
  };

  aes_cleanup();
  if (output_size_required < 0)
    bzero(orig_output, output_size);
  return output_size_required;
}


/* destroy a key, but don't free it */
void
aes_key_zero(aes_key *key) {
  bzero(key, sizeof(*key));
}

/* destroy a key */
void
aes_key_destroy(aes_key *key) {
  aes_key_zero(key);
  free(key);
}

/* schedule a key */
aes_key *aes_schedule(const byte          *input,
		            unsigned int   nbits,
		            aes_key       *buffer,
   		            aes_direction  direction) {
  if ((nbits != 128
       && nbits != 192
       && nbits != 256)
      || (direction != AES_DIRECTION_ENCRYPT
	  && direction != AES_DIRECTION_DECRYPT))
    return NULL;

  /* allocate memory for the key */
  if (buffer == NULL) {
    buffer = calloc(sizeof(*buffer), 1);
    if (buffer == NULL) return NULL;
  } else {
    aes_key_zero(buffer);
  }
  
  /* initialize nrounds */
  int nrounds = nbits / 32 + 6;
  buffer->_n_middle_rounds = nrounds - 1;

  /* to do: more in C */
  if (direction == AES_DIRECTION_ENCRYPT) {
    aes_schedule_core(input,
		      nbits,
		      &buffer->_subkey[0],
		      direction,
		      48);
  } else {
    aes_schedule_core(input,
		      nbits,
		      /* output keys in reverse order */
		      &buffer->_subkey[nrounds],
		      direction,
		      (nbits == 192) ? 0 : 32);
  }

  return buffer;
}
