/* Constant-time SSSE3 AES core API.
 * version 0.1
 *
 * By Mike Hamburg (Stanford University), May 2009
 * Public domain.
 *
 * This API is still under development, and so is subject to change.
 * (For example, the CBC code should probably take sizes in bytes...)
 * It also might be buggy.  Caveat emptor.
 */

#include <tmmintrin.h>
#include <errno.h>

typedef unsigned char byte;

/* Keys are 256 bytes long and aligned to 16 bytes.  They include a
 * round counter and 10-14 round subkeys.
 *
 * Each key is only good for encryption or for decryption.
 *
 * Labelled volatile so that the compiler won't load the key when told
 * not to.
 */
typedef struct {
  u_int64_t        _n_middle_rounds;
  u_int64_t        _zero_pad;
  volatile __m128i _subkey[15];
} aes_key __attribute__((__aligned__(16)));

/* Directions for aes_schedule */
typedef enum {
  AES_DIRECTION_ENCRYPT,
  AES_DIRECTION_DECRYPT
} aes_direction;

/* Padding modes */
typedef enum {
  AES_PADDING_NULL        = 0,
  AES_PADDING_DES         = 1,
  AES_PADDING_PKCS5       = 2,
  AES_CIPHERTEXT_STEALING = 3,
  AES_PADDING_NONE        = 4
} aes_padding;

/* Modes of operation */
typedef enum {
  AES_MODE_ECB,
  AES_MODE_CBC,
  AES_MODE_CTR_64,
  AES_MODE_CTR_128,
  /* unimplemented */
  AES_MODE_CFB,
  AES_MODE_OFB,
  AES_MODE_OCB
} aes_mode;

#define block_size 16

/* Schedule out an AES key with (nbits) bits.
 *
 * nbits must be 128, 192 or 256.
 *
 * direction must be AES_DIRECTION_ENCRYPT or AES_DIRECTION_DECRYPT
 *
 * If you output != null, the key will be scheduled at output, and
 * output will be returned.  Note that the output must be a
 * 16-byte-aligned, 256-byte structure.  This is useful for reusing
 * key storage.
 */
aes_key *aes_schedule(const byte          *input,
		            unsigned int   nbits,
		            aes_key       *buffer,
		            aes_direction  direction);

int64_t
aes_encrypt(const byte      *input,
	    u_int64_t        input_size,
	    byte            *output,
	    u_int64_t        output_size,
	    const aes_key   *key,
	    const byte      *iv,
	    aes_mode         mode,
	    aes_padding      padding);
	    
int64_t
aes_decrypt(const byte      *input,
	    u_int64_t        input_size,
	    byte            *output,
	    u_int64_t        output_size,
	    const aes_key   *key,
	    const byte      *iv,
	    aes_mode         mode,
	    aes_padding      padding);




/* Zero over and free an AES key.
 * Equivalent to { bzero(key, sizeof(*key)); free(key); }
 */
void aes_key_destroy(aes_key *key);

/* Zero the key after using it. */
void aes_key_zero(aes_key *key);

