/* AES implementation internals */

#include <strings.h>
#include "aes.h"

/* Blocks are 128 bits long.  They may be aligned or unaligned. */
typedef __m128i block;
typedef __m128i unaligned_block __attribute__((__aligned__(1)));

/* Structure of the AES constant tables */
typedef struct {
  __m128i lo, hi;
} transform;

typedef struct {
  __m128i u, t;
} sbox_output;

typedef __m128i permutation, lookup, mask, add_const;

typedef struct {
  /* encryption core */
  mask        s0F;       /* repeated 0x0F */
  transform   ipt;       /* input transform for encryption */
  lookup      inv;       /* invert in GF(2^4) */
  lookup      inva;      /* invert * constant "a" */
  sbox_output sb1;       /* sbox * 1 */
  sbox_output sb2;       /* sbox * 2 */
  sbox_output sbo;       /* sbox, deskew */

  /* mix columns permutations, conjugated by shiftrows */
  permutation mc_forward[4];
  permutation mc_backward[4];

  /* shift rows permutations */
  permutation sr[4];

  /* key schedule */
  lookup      rcon;      /* round constants */
  add_const   s63;       /* 0x63 in altered basis */
  transform   opt;       /* output transform */
  transform   deskew;    /* undo sbox skew */
  transform   dks[4];    /* decryption matrix constants */

  /* decryption core */
  transform   dipt;      /* input transform for decryption */
  sbox_output dsb9;      /* sbox output *9 */
  sbox_output dsbd;      /* sbox output *d */
  sbox_output dsbb;      /* sbox output *b */
  sbox_output dsbe;      /* sbox output *e */
  sbox_output dsbo;      /* sbox output final round */

  /* modes */
  mask        ctr_mask;  /* helper for long increment */
  mask        ctr_one;   /* 0,1  */
  mask        ocb_mask;  /* used by ocb_double */

  permutation reverse_id;
} __attribute((__aligned__(16))) aes_consts;

/* Assembly glue. */

/* Local register variables.  For performance reasons, we leave some
 * important lookup tables in registers.  These macros load the
 * registers.
 */
#define aes_use_locals()		    \
  register __m128i _x09 asm("%xmm9");       \
  register __m128i _x10 asm("%xmm10");      \
  register __m128i _x11 asm("%xmm11");      \
  register __m128i _x12 asm("%xmm12");      \
  register __m128i _x13 asm("%xmm13");      \
  register __m128i _x14 asm("%xmm14");      \
  register __m128i _x15 asm("%xmm15");      \
  register const aes_consts *_base asm("%r10"); \
  asm ("call _aes_preheat"                  \
      : "=x"(_x09), "=x"(_x10), "=x"(_x11), \
	"=x"(_x12), "=x"(_x13), "=x"(_x14), \
	"=x"(_x15), "=r"(_base));

/* The local variables */
#define AES_LOCALS                          \
  "x"(_x09), "x"(_x10), "x"(_x11),          \
  "x"(_x12), "x"(_x13), "x"(_x14),          \
  "x"(_x15), "r"(_base)

/* Erase sensitive registers.  Call this at the end of encryption
 * routines.  Ideally, this could be tail-called, but we'll settle for
 * an ordinary call.
 */
#define aes_cleanup()			    \
  asm volatile("call _aes_cleanup"          \
	       : /* no output */	    \
	       : AES_LOCALS		    \
	       : "%xmm0", "%xmm1", "%xmm2", \
		 "%xmm3", "%xmm4", "%xmm5", \
		 "%xmm6", "%xmm7", "%xmm8"  \
	       );

/* AES encryption core. */
#define aes_encrypt_core(_input, _key) ({   \
      register block _io asm("%xmm0");	    \
      _io = _input;			    \
  asm ("call _aes_encrypt_core"             \
       : "=x"(_io)			    \
       : "x"(_io),  AES_LOCALS, "d"(key)    \
       : "%xmm1", "%xmm2", "%xmm3",         \
	 "%xmm4", "%r9", "%r11", "%rax");   \
  _io;					    \
    })

/* Decryption core. */
#define aes_decrypt_core(_input, _key) ({		\
      register block _io asm("%xmm0");			\
      _io = _input;					\
      asm ("call _aes_decrypt_core"			\
       : "=x"(_io)					\
       : "x"(_io),  AES_LOCALS, "d"(key)		\
       : "%xmm1", "%xmm2", "%xmm3",			\
	 "%xmm4", "%xmm5", "%r9", "%r11", "%rax");	\
      _io;						\
    })

/* Schedule */
aes_key *aes_schedule_core(const byte       *input,
			   unsigned int      nbits,
			   volatile __m128i *buffer,
			   aes_direction     direction,
			   int               rotation_offset)
			   __attribute__((sysv_abi));


/* Load _n bytes from _p into a vector register %xmm0 and return it.
 * Apply the given padding mode.
 */
#define loadup(_p, _n, _chain, _mode) ({		\
      register block _io asm("%xmm0");			\
      _io = _chain;					\
      asm("call _unaligned_partial_load"		\
	  : "=x"(_io)					\
	  : "c"(_n), "D"(_p), "r"(_base),		\
	    "x"(_io), "a"(_mode)			\
	  : "%r8", "%xmm1", "%xmm2", "%xmm3", "%xmm4"	\
	  );						\
      _io;})

/* Store _n bytes in _x=%xmm0 to _p */
#define storeup(_p, _n, _x) {				\
  register block _io asm("%xmm0");			\
  _io = _x;						\
  asm("call _unaligned_partial_store"			\
      : /* no output */					\
      : "c"(_n), "S"(_p), "r"(_base), "x"(_io)		\
      : "%rdi", "%xmm1", "%xmm2", "%xmm3", "%xmm4",	\
      "%xmm5", "memory");				\
  }

#define check_padding(_x, _mode) ({		\
  int _ret;					\
  register block _io asm("%xmm0");		\
  _io = _x;					\
  asm("call _check_padding"			\
      : "=c"(_ret)				\
      : "r"(_base), "x"(_io), "a"(_mode)	\
      :	"%xmm1", "%xmm2", "%xmm3", "%r8"	\
      );					\
  _ret;})

/* Abbreviated intrinsics. */
#define xor       _mm_xor_si128
#define and       _mm_and_si128
#define or        _mm_or_si128
#define andn      _mm_andnot_si128
#define merge(_m,_x,_y) xor(and(_x,_m),andn(_y,_m))
#define psubq     _mm_sub_epi64
#define paddb     _mm_add_epi8
#define paddq     _mm_add_epi64
#define pcmpeqd   _mm_cmpeq_epi32
#define pcmpeqb   _mm_cmpeq_epi8
#define zero      _mm_setzero_si128
#define not(_x)   pcmpeqb(_x, zero())
#define pslldq    _mm_slli_si128
#define psrad     _mm_srai_epi32
#define pslld     _mm_slli_epi32
#define pshufb    _mm_shuffle_epi8
#define pcmpgtb   _mm_cmpgt_epi8
#define loadu(_x) _mm_loadu_si128((const unaligned_block *) (_x))
#define storeu(_p, _x) _mm_storeu_si128((unaligned_block *) (_p), _x)
#define pshufd(_x,_a,_b,_c,_d) \
  _mm_shuffle_epi32(_x, _a+4*_b+16*_c+64*_d)
