#include <tmmintrin.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <stdlib.h>
#include "aes.h"

#define die(msg, args...) { \
  fprintf(stderr, "test: error: " msg "\n", ##args); \
  exit(1); \
}

/* print a usage message and exit */
static void
usage() {
  fprintf(stderr,
    "Usage: ./test options ...\n"
    "\n"
    "./test -b benchmarks the cipher\n"
    "./test -B benchmarks the key schedule\n"
    "./test -i gives the output for a given input\n"
    "./test -i foo -o bar runs a known-answer test\n"
    "./test without any of these options tests that\n"
    "       decryption undoes encryption\n"
    "\n"
    "Options:\n"
    "  -b,--benchmark nruns\n"
    "  -B,--benchmark-key-schedule nruns\n"
    "  -k,--key-size 128|192|256\n"
    "  -n,--iterations iterations\n"
    "  -s,--packet-size size\n"
    "  -m,--mode cbc|ecb|ctr|ctr64|cfb|ofb|ocb\n"
    "  -p,--padding none|null|pkcs5|des|cts\n"
    "  -d,--decrypt\n"
    "\n"
    "  -i,--input hex\n"
    "  -o,--output hex\n"
    "  -I,--iv hex\n"
    "  -K,--key hex\n"
    "\n"
    );
  exit(1);
}

static int
compareu(const void *a, const void *b) {
  u_int64_t aa = *(u_int64_t *) a, bb = *(u_int64_t *) b;
  if (aa > bb) return 1;
  if (bb > aa) return -1;
  return 0;
}

static double
median(u_int64_t *meas, int n) {
  qsort((void *) meas, n, sizeof(u_int64_t), compareu);
  if (n & 1) {
    return (double) meas[n/2];
  } else {
    return (meas[n/2] + meas[n/2+1])/2.0;
  }
}

static void hex_print(FILE *out, const char *pre, int len, const byte *x) {
  fprintf(out, "%s", pre);
  int i;
  for (i=0; i<len; i++) {
    fprintf(out, "%02x", (int) x[i]);
  }
  fprintf(out, "\n");
}

/* read timestamp for benchmarks */
static inline u_int64_t rdtsc(void)
{
  u_int64_t hi, lo;
  asm volatile ("rdtsc" : "=a"(lo), "=d"(hi));
  return lo | (hi << 32);
}


/* return the mode with the given name */
static aes_mode
string_to_mode(char *m) {
  int i;
  struct {
    const char *name;
    aes_mode mode;
  } modes[] = { {"ecb",   AES_MODE_ECB },
		{"cbc",   AES_MODE_CBC },
		{"ctr",   AES_MODE_CTR_128 },
		{"ctr64", AES_MODE_CTR_64 },
		{"ofb",   AES_MODE_OFB },
		{"cfb",   AES_MODE_CFB },
		{"ocb",   AES_MODE_OCB },
		{NULL,    -1 }};

  for (i=0;;i++) {
    if (modes[i].name == NULL) {
      die("Invalid mode \"%s\"\n", m);
    } else if (strcasecmp(modes[i].name, m) == 0) {
      return modes[i].mode;
    }
  }
}

static byte *
parse_hex(const char *hex, int *length) {
  int i, len = strlen(hex);
  byte h=0;
  byte *out = NULL;
  
  int mallen = 32;
  
  if (len % 2 != 0) goto fail;
  if (len / 2 > mallen) mallen = len / 2;
  if (length != NULL) *length = len/2;
  out = calloc(mallen, 1);
  if (out == NULL) goto fail;

  for (i=0; i<len; i++) {
    char c = hex[i];
    if (c >= '0' && c <= '9') {
      h += c - '0';
    } else if (c >= 'A' && c <= 'F') {
      h += 10 + c - 'A';
    } else if (c >= 'a' && c <= 'f') {
      h += 10 + c - 'a';
    } else {
      goto fail;
    }
    if (i % 2 == 1) {
      out[i/2] = h;
      h = 0;
    } else {
      h *= 16;
    }
  }
  
  return out;
  
fail:
  if (out != NULL) free(out);
  if (length != NULL) *length = 0;
  return NULL;
}

/* return the padding with the given name */
static aes_padding
string_to_padding(char *m) {
  int i;
  struct {
    const char *name;
    aes_padding pad;
  } pads[] = { {"null",   AES_PADDING_NULL  },
	       {"des",    AES_PADDING_DES   },
	       {"pkcs5",  AES_PADDING_PKCS5 },
	       {"cts",    AES_CIPHERTEXT_STEALING },
	       {"none",   AES_PADDING_NONE },
	       {NULL,    -1 }};

  for (i=0;;i++) {
    if (pads[i].name == NULL) {
      die("Invalid padding \"%s\"", m);
    } else if (strcasecmp(pads[i].name, m) == 0) {
      return pads[i].pad;
    }
  }
}

/* return a suitable padding for the given mode */
static aes_padding
padding_for_mode(aes_mode m) {
  switch(m) {
  case AES_MODE_ECB:
  case AES_MODE_CBC:
    return AES_CIPHERTEXT_STEALING;

  case AES_MODE_CTR_64:
  case AES_MODE_CTR_128:
  case AES_MODE_CFB:
  case AES_MODE_OFB:
  case AES_MODE_OCB:
  default:
    return AES_PADDING_NONE;
  };
}


int is_prf_mode(aes_mode m) {
  switch(m) {
  case AES_MODE_ECB:
  case AES_MODE_CBC:
  case AES_MODE_OCB:
    return 0;

  default:
    return 1;
  }
}

static off_t
verify_encryption(const byte *pt1,
		  byte *ct, byte *pt2,
		  size_t p, size_t q,
		  aes_key *key, aes_key *dkey,
		  const byte *iv,
		  aes_mode mode,
		  aes_padding padding) {
  
  off_t ret = aes_encrypt(pt1, p, ct, q,
			  key, iv, mode, padding);
  if (ret < 0) {
    fprintf(stderr, "Encryption error: %s\n", strerror(-ret));
    exit(1);
  }

  off_t ret2 = aes_decrypt(ct, ret, pt2, q,
			   is_prf_mode(mode) ? key : dkey,
			   iv, mode, padding);
  if (ret2 < 0) {
    fprintf(stderr, "Decryption error: %s\n", strerror(-ret));
    exit(1);
  }

  if (ret2 != p) {
    fprintf(stderr, "Decryption returns wrong size (%ld vs %ld)\n",
	    ret2, p);
    exit(1);
  }

  size_t i;
  for (i=0; i<p; i++) {
    if (pt1[i] != pt2[i]) {
      fprintf(stderr, "Decryption wrong in byte %ld of %ld\n",
	      i, p);
      exit(1);
    }
  }

  return ret;
}

#define requires_argument() { \
  if (i >= argc - 1) {     \
    die("%s requires an argument", argv[i]); \
  } \
  i++; \
}

#define requires_integer(_x) { \
  requires_argument();         \
  _x = atoi(argv[i]);          \
}

#define argument(short, long) \
  if (strcmp(arg, short) == 0 || strcmp(arg, long) == 0)

int main(int argc, char **argv) {
  int key_size = 0;
  int packet_size = 576;
  int64_t iterations = 1000;
  int64_t i;
  int benchmark = 0;
  char *mode_str = "cbc";
  u_int64_t start;
  int64_t nruns = 1;
  aes_direction direction = AES_DIRECTION_ENCRYPT;
  aes_mode    mode = AES_MODE_CBC;
  aes_padding padding = -1;
  
  int k_len = 0;
  byte *ikey = parse_hex(
    "000102030405060708090a0b0c0d0e0"
    "101112131415161718191a1b1c1d1e1",
    NULL
  );
  
  byte *iv = parse_hex(
    "00000000000000000000000000000000",
    NULL
  );
  
  int i_len = 0;
  byte *input = NULL;
  int o_len = 0;
  byte *output = NULL;
	
  /* parse arguments */
  for (i=1; i<argc; i++) {
    char *arg = argv[i];
    
    argument("-k", "--key-size") {
      requires_integer(key_size);
      switch(key_size) {
	case 128: case 192: case 256: break;
	default:
	  die("%s is not a valid key size", argv[i+1]);
      };
    } else argument("-s", "--packet-size") {
      requires_integer(packet_size);
      if (packet_size < 0)
	die("packet_size must be a nonnegative integer");
    } else argument("-n", "--iterations") {
      requires_integer(iterations);
      if (iterations < 0)
	die("iterations must be a nonnegative integer");
    } else argument("-b", "--benchmark") {
      requires_integer(nruns);
      benchmark = 1;
    } else argument("-B", "--benchmark-key-schedule") {
      requires_integer(nruns);
      if (nruns < 0)
	die("nruns must be a nonnegative integer");
      benchmark = 2;
    } else argument("-m", "--mode") {
      requires_argument();
      mode = string_to_mode(mode_str = argv[i]);
    } else argument("-p", "--padding") {
      requires_argument();
      padding = string_to_padding(argv[i]);
    } else argument("-d", "--decrypt") {
      direction = AES_DIRECTION_DECRYPT;
    } else argument("-I", "--iv") {
      requires_argument();
      int l;
      
      free(iv);
      iv = parse_hex(argv[i], &l);
      if (iv == NULL) die("invalid IV");
      if (l > 16) die("IV too long");
    
    } else argument("-K", "--key") {
      requires_argument();
      
      free(ikey);
      ikey = parse_hex(argv[i], &k_len);
      if (ikey == NULL) die("invalid key");
      if (k_len > 32) die("key too long");
    
    } else argument("-i", "--input") {
      requires_argument();
      
      if (input != NULL) free(input);
      input = parse_hex(argv[i], &i_len);
      if (input == NULL) die("invalid input");
      
    } else argument("-o", "--output") {
      requires_argument();
      
      if (output != NULL) free(output);
      output = parse_hex(argv[i], &o_len);
      if (output == NULL) die("invalid output");
      
    } else {
      //die("invalid argument %s", argv[i]);
      usage();
    }
  }
  
  if (padding == -1)
    padding = padding_for_mode(mode);
  
  if (key_size == 0) {
    if (k_len <= 16) {
      key_size = 128;
    } else if (k_len <= 24) {
      key_size = 192;
    } else if (k_len <= 32) {
      key_size = 256;
    } else {
      die("key is too long (%d bytes)!", k_len);
    }
  }
  
  u_int64_t *runs = calloc(sizeof(*runs), nruns);

  if (output != NULL && input == NULL) {
    die("output specified but not input");
  }
  
  aes_key
    *key  = aes_schedule((byte *) ikey, key_size, NULL,
			 AES_DIRECTION_ENCRYPT),

    *dkey = aes_schedule((byte *) ikey, key_size, NULL,
			 AES_DIRECTION_DECRYPT);
  
  if (input != NULL) {
    if (benchmark != 0) die("-i and -b/-B are incompatible");
    
    byte *output2 = malloc(i_len + 2*block_size);
    off_t rsz;
    if (direction == AES_DIRECTION_ENCRYPT) {
      rsz = aes_encrypt(input, i_len,
			output2, i_len + 2*block_size,
			key, iv, mode, padding);
    } else {
      rsz = aes_decrypt(input, i_len,
			output2, i_len + 2*block_size,
			is_prf_mode(mode) ? key : dkey, iv, mode, padding);
    }
    
    if (rsz < 0) {
      die("KAT error: %s", strerror(-rsz));
    } else if (output != NULL) {
      int fail = 0;
      if (rsz != o_len) {
	fail = 1;
	fprintf(stderr, "test: error: KAT gives length %ld instead of %d\n",
		rsz, o_len);
      } else {
	for (i=0; i<o_len; i++) {
	  if (output[i] != output2[i]) {
	    fail=1;
	    fprintf(stderr, "test: error: KAT test failed\n");
	    break;
	  }
	}
      }
      
      if (fail) {
	hex_print(stderr, "expected output: ", o_len, output);
	hex_print(stderr, "actual   output: ", rsz,   output2);
      }
      
      return fail;
    } else {
      hex_print(stdout, "", rsz, output2);
    }
    
    free(output2);
    return 0;
  }
  
  if (benchmark == 2) {
    key = aes_schedule((byte *) ikey, key_size, NULL, direction);

    start = rdtsc();
    for (int j=0; j<nruns; j++) {
      for (i=0; i<iterations; i++) {
	aes_schedule((byte *) ikey, key_size, key, direction);
      }
      runs[j] = rdtsc() - start;
    }
    for (int j=nruns-1; j>0; j--) {
      runs[j] -= runs[j-1];
    }
    runs[0] -= start;
    
    double cpi = median(runs, nruns) / iterations;
      printf("aes-%d (schedule %s): %0.2f cpi\n",
	key_size,
	(direction == AES_DIRECTION_ENCRYPT) ? "enc" : "dec",
	cpi);
      
    return 0;
  }

  /* construct the plaintext buffers */
  byte *pt1 = malloc(packet_size+5),
    *pt2 = malloc(packet_size+7),
    *ct  = malloc(packet_size + 2*block_size + 12);
  const char *pttxt = "Sphinx of black quartz, judge my vow!\n";
  int len = strlen(pttxt);
  for (size_t i = 0; i < packet_size; i += len) {
    if (i+len >= packet_size)
      memcpy(pt1+i, pttxt, packet_size-i);
    else
      memcpy(pt1+i, pttxt, len);
  }
      
  verify_encryption(pt1, ct, pt2,
		    packet_size, packet_size+2*block_size,
		    key, dkey, iv, mode, padding);

  
  if (benchmark == 1) {
    switch(direction) {
      case AES_DIRECTION_ENCRYPT:
	start = rdtsc();
	for (int j=0; j<nruns; j++) {
	  for (i=0; i<iterations; i++) {
	    off_t ret = aes_encrypt(pt1, packet_size,
				    ct, packet_size + 2*block_size,
				    key, iv, mode, padding);
	    assert(ret >= 0);
	  }
	  runs[j] = rdtsc();
	}
	break;
    
      case AES_DIRECTION_DECRYPT: {
	aes_key *key2 = is_prf_mode(mode) ? key : dkey;
	start = rdtsc();
	for (int j=0; j<nruns; j++) {
	  for (i=0; i<iterations; i++) {
	    off_t ret = aes_decrypt(ct, packet_size,
				    pt2, packet_size + 2*block_size,
				    key2, iv, mode, padding);
	    assert(ret >= 0);
	  }
	  runs[j] = rdtsc();
	}
	break;
      }

    default: start = rdtsc();
    }
    for (int j=nruns-1; j>0; j--) {
      runs[j] -= runs[j-1];
    }
    runs[0] -= start;
  
    if (packet_size) {
      double cpb = median(runs,nruns) / (iterations * packet_size);
      printf("aes-%d-%s (%s %4d B): %0.2f cpb\n",
	     key_size, mode_str,
	     (direction == AES_DIRECTION_ENCRYPT) ? "enc" : "dec",
	     packet_size, cpb);
    } else {
      printf("aes-%d-%s (%s calibrate): %0.2f cpi\n",
	     key_size, mode_str,
	     (direction == AES_DIRECTION_ENCRYPT) ? "enc" : "dec",
	     median(runs, nruns) / iterations);
    }
  }
  aes_key_destroy(key);
  aes_key_destroy(dkey);
  //free(pt1);
  //free(pt2);
  //free(ct);
  
  return 0;
}
